import os
import urllib
from datetime import datetime
from enum import Enum

import boto3
from boto3.s3.transfer import TransferConfig
from pydantic import BaseModel, Field
from rich.progress import (
    BarColumn,
    DownloadColumn,
    Progress,
    TextColumn,
    TimeRemainingColumn,
    TransferSpeedColumn,
)

from .config import Config
from .httputils import flow360ApiGet

_s3_config = TransferConfig(multipart_threshold=1024 * 25, max_concurrency=50,
                            multipart_chunksize=1024 * 25, use_threads=True)


class _S3Action(Enum):
    Uploading = "↑"
    Downloading = "↓"


def _get_progress(action: _S3Action):
    col = TextColumn(f"[bold green]{_S3Action.Downloading.value}") if action == _S3Action.Downloading else TextColumn(
        f"[bold red]{_S3Action.Uploading.value}")
    return Progress(
        col,
        TextColumn("[bold blue]{task.fields[filename]}"),
        BarColumn(),
        "[progress.percentage]{task.percentage:>3.1f}%",
        "•",
        DownloadColumn(),
        "•",
        TransferSpeedColumn(),
        "•",
        TimeRemainingColumn(),
    )


class _UserCredential(BaseModel):
    access_key_id: str = Field(alias="accessKeyId")
    expiration: datetime
    secret_access_key: str = Field(alias="secretAccessKey")
    session_token: str = Field(alias="sessionToken")


class _S3STSToken(BaseModel):
    cloud_path: str = Field(alias="cloudpath")
    user_credential: _UserCredential = Field(alias="userCredentials")

    def get_bucket(self):
        r = urllib.parse.urlparse(self.cloud_path)
        return r.netloc

    def get_s3_key(self):
        r = urllib.parse.urlparse(self.cloud_path)
        return r.path[1:]

    def get_client(self):
        return boto3.client(
            's3',
            region_name=Config.AWS_REGION,
            aws_access_key_id=self.user_credential.access_key_id,
            aws_secret_access_key=self.user_credential.secret_access_key,
            aws_session_token=self.user_credential.session_token,
        )

    def is_expired(self):
        return (self.user_credential.expiration.replace(tzinfo=None) - datetime.utcnow()).total_seconds() > 300


class S3TransferType(Enum):
    VolumeMesh = "VolumeMesh"
    SurfaceMesh = "SurfaceMesh"
    Case = "Case"
    Studio = "Studio"

    def _get_grant_url(self, resource_id, file_name: str) -> str:
        if self is S3TransferType.VolumeMesh:
            return f"volumemeshes/{resource_id}/file?filename={file_name}"
        if self is S3TransferType.SurfaceMesh:
            return f"surfacemeshes/{resource_id}/file?filename={file_name}"
        if self is S3TransferType.Case:
            return f"cases/{resource_id}/file?filename={file_name}"
        if self is S3TransferType.Studio:
            return f"volumemeshes/{resource_id}/file?filename={file_name}"


    def upload_file(self, resource_id: str, remote_file_name: str, file_name: str):
        if not os.path.exists(file_name):
            print('mesh file {0} does not Exist!'.format(file_name))
            raise FileNotFoundError()
        with _get_progress(_S3Action.Uploading) as progress:
            task_id = progress.add_task("upload", filename=os.path.basename(file_name),
                                        total=os.path.getsize(file_name))

            def _call_back(bytes_in_chunk):
                progress.update(task_id, advance=bytes_in_chunk)

            token = self._get_s3_sts_token(resource_id, remote_file_name)
            token.get_client().upload_file(Bucket=token.get_bucket(),
                                           Filename=file_name,
                                           Key=token.get_s3_key(),
                                           Callback=_call_back,
                                           Config=_s3_config)

    def download_file(self, resource_id: str, remote_file_name: str, to_file: str):
        token = self._get_s3_sts_token(resource_id, remote_file_name)
        client = token.get_client()
        meta_data = client.head_object(Bucket=token.get_bucket(), Key=token.get_s3_key())
        with _get_progress(_S3Action.Downloading) as progress:
            progress.start()
            task_id = progress.add_task("download", filename=os.path.basename(remote_file_name),
                                        total=meta_data.get('ContentLength', 0))

            def _call_back(bytes_in_chunk):
                progress.update(task_id, advance=bytes_in_chunk)

            os.makedirs(os.path.dirname(os.path.abspath(to_file)), exist_ok=True)

            client.download_file(Bucket=token.get_bucket(),
                                 Filename=to_file,
                                 Key=token.get_s3_key(),
                                 Callback=_call_back
                                 )

    def _get_s3_sts_token(self, resource_id: str, file_name: str) -> _S3STSToken:
        session_key = f"{resource_id}:{self.value}:{file_name}"
        if session_key not in _s3_sts_tokens or _s3_sts_tokens[session_key].is_expired():
            method = self._get_grant_url(resource_id, file_name)
            resp = flow360ApiGet(method)
            token = _S3STSToken.parse_obj(resp)
            _s3_sts_tokens[session_key] = token
        return _s3_sts_tokens[session_key]


_s3_sts_tokens: [str, _S3STSToken] = {}
